
#include <dinput.h>
#include "DirectXVideoOutputDevice.h"
#include "DirectInputDevice.h"
#include "DirectInputSignal.h"

extern DirectXVideoOutputDevice* currentDevice;

INT32 initDirectInput();
void releaseDirectInput();

struct JOYAXESDATA
{
    LONG lMaxDeltaX;
    LONG lCenterX;
    LONG lMaxDeltaY;
    LONG lCenterY;
    LONG lMaxDeltaZ;
    LONG lCenterZ;

};

struct DirectInputDeviceClass
{

    IDirectInput8* m_pDI;
    BOOL m_bLastPollSuccessful;
    IDirectInputDevice8* m_pdiKeyboard;
    char m_KeyboardBuffer[256];

    IDirectInputDevice8* m_pdiJoysticks[10];
    char m_JoystickNames[10][MAX_PATH];
    JOYAXESDATA m_JoystickAxes[10];
    DIJOYSTATE2 m_JoystickStates[10];
    INT32 m_iJoystickCount;
};

//currently only one global input instance supported
DirectInputDeviceClass inputInstance;

INT32 g_iJoyButtonIDStart = 0x0000;
INT32 g_iJoyAxesIDStart = 0x1000;

UINT64 DEFAULT_KEYS_PLAYER_ONE[] = {
        DIK_1, DIK_2, DIK_3, DIK_4, DIK_5, DIK_6, DIK_7, DIK_8, DIK_9,
        DIK_DELETE, DIK_0, DIK_RETURN, DIK_LCONTROL, DIK_Z, DIK_X,
        DIK_UP, 0, DIK_RIGHT, 0, DIK_DOWN, 0, DIK_LEFT, 0
};

UINT64 DEFAULT_KEYS_PLAYER_TWO[] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

UINT64 DEFAULT_KEYS_ECS_KEYBOARD[48] = {
    DIK_LEFT, DIK_COMMA, DIK_N, DIK_V, DIK_X, DIK_SPACE,
    DIK_PERIOD, DIK_M, DIK_B, DIK_C, DIK_Z, DIK_DOWN,
    DIK_SEMICOLON, DIK_K, DIK_H, DIK_F, DIK_S, DIK_UP,
    DIK_P, DIK_I, DIK_Y, DIK_R, DIK_W, DIK_Q,
    DIK_ESCAPE, DIK_9, DIK_7, DIK_5, DIK_3, DIK_1,
    DIK_0, DIK_8, DIK_6, DIK_4, DIK_2, DIK_RIGHT,
    DIK_RETURN, DIK_O, DIK_U, DIK_T, DIK_E, DIK_LCONTROL,
    DIK_L, DIK_J, DIK_G, DIK_D, DIK_A, DIK_LSHIFT
};

UINT64 DEFAULT_KEYS_DIRECTOR[6] = {
    DIK_ESCAPE, DIK_F1, DIK_F5, DIK_F6, DIK_F9, DIK_F12
};

char* g_lpszPOVDescriptionKeys[4] = { "POVUpDescription",
    "POVRightDescription",
    "POVDownDescription",
    "POVLeftDescription" };

DirectInputSignal* DEFAULT_INPUT_SIGNALS_PLAYER_ONE[23];
DirectInputSignal* DEFAULT_INPUT_SIGNALS_PLAYER_TWO[23];
DirectInputSignal* DEFAULT_INPUT_SIGNALS_ECS_KEYBOARD[48];
DirectInputSignal* DEFAULT_INPUT_SIGNALS_DIRECTOR[6];

DirectInputDevice::DirectInputDevice()
{
    initDirectInput();

    //initialize the default player one input signals
    DIPROPSTRING propString;
    propString.diph.dwSize = sizeof(DIPROPSTRING);
    propString.diph.dwHeaderSize = sizeof(DIPROPHEADER);
    propString.diph.dwHow = DIPH_BYOFFSET;
    for (INT32 i = 0; i < 23; i++) {
        if (DEFAULT_KEYS_PLAYER_ONE[i] == 0)
            DEFAULT_INPUT_SIGNALS_PLAYER_ONE[i] = NULL;
    
        propString.diph.dwObj = DEFAULT_KEYS_PLAYER_ONE[i];
        inputInstance.m_pdiKeyboard->GetProperty(DIPROP_KEYNAME,
                &propString.diph);

        DEFAULT_INPUT_SIGNALS_PLAYER_ONE[i] = new DirectInputSignal(0,
                DEFAULT_KEYS_PLAYER_ONE[i]);
    }

    //initialize the player two input signals
    for (i = 0; i < 23; i++) {
        if (DEFAULT_KEYS_PLAYER_TWO[i] == 0)
            DEFAULT_INPUT_SIGNALS_PLAYER_TWO[i] = NULL;
    
        propString.diph.dwObj = DEFAULT_KEYS_PLAYER_TWO[i];
        inputInstance.m_pdiKeyboard->GetProperty(DIPROP_KEYNAME,
                &propString.diph);

        DEFAULT_INPUT_SIGNALS_PLAYER_TWO[i] = new DirectInputSignal(0,
                DEFAULT_KEYS_PLAYER_TWO[i]);
    }

    //initialize the ecs keyboard input signals
    for (i = 0; i < 48; i++) {
        if (DEFAULT_KEYS_ECS_KEYBOARD[i] == 0)
            DEFAULT_INPUT_SIGNALS_ECS_KEYBOARD[i] = NULL;
   
        propString.diph.dwObj = DEFAULT_KEYS_ECS_KEYBOARD[i];
        inputInstance.m_pdiKeyboard->GetProperty(DIPROP_KEYNAME,
                &propString.diph);

        DEFAULT_INPUT_SIGNALS_ECS_KEYBOARD[i] = new DirectInputSignal(0,
                DEFAULT_KEYS_ECS_KEYBOARD[i]);
    }

    //initialize the control input signals
    for (i = 0; i < 6; i++) {
        if (DEFAULT_KEYS_DIRECTOR[i] == 0)
            DEFAULT_INPUT_SIGNALS_DIRECTOR[i] = NULL;
   
        propString.diph.dwObj = DEFAULT_KEYS_DIRECTOR[i];
        inputInstance.m_pdiKeyboard->GetProperty(DIPROP_KEYNAME,
                &propString.diph);

        DEFAULT_INPUT_SIGNALS_DIRECTOR[i] = new DirectInputSignal(0,
                DEFAULT_KEYS_DIRECTOR[i]);
    }

    releaseDirectInput();
}

void DirectInputDevice::poll()
{
    //update the keyboard buffer
    if (FAILED(inputInstance.m_pdiKeyboard->GetDeviceState(
            sizeof(inputInstance.m_KeyboardBuffer),
            (LPVOID)&inputInstance.m_KeyboardBuffer)))
    {
        //try to reacquire the device before giving up on the keyboard
        if (FAILED(inputInstance.m_pdiKeyboard->Acquire())) {
            inputInstance.m_bLastPollSuccessful = FALSE;
            return;
        }

        if (FAILED(inputInstance.m_pdiKeyboard->GetDeviceState(
                    sizeof(inputInstance.m_KeyboardBuffer),
                    (LPVOID)&inputInstance.m_KeyboardBuffer)))
        {
            inputInstance.m_bLastPollSuccessful = FALSE;
            return;
        }
    }

    //update the joystick states
    for (INT32 i = 0; i < inputInstance.m_iJoystickCount; i++) {
        if (FAILED(inputInstance.m_pdiJoysticks[i]->Poll()))
        {
            if (FAILED(inputInstance.m_pdiJoysticks[i]->Acquire())) {
                inputInstance.m_bLastPollSuccessful = FALSE;
                return;
            }
    
            if (FAILED(inputInstance.m_pdiJoysticks[i]->Poll())) {
                inputInstance.m_bLastPollSuccessful = FALSE;
                return;
            }
        }

        if (FAILED(inputInstance.m_pdiJoysticks[i]->GetDeviceState(
                sizeof(DIJOYSTATE2), &inputInstance.m_JoystickStates[i])))
        {
            inputInstance.m_bLastPollSuccessful = FALSE;
            return;
        }
    }

    inputInstance.m_bLastPollSuccessful = TRUE;
}

BOOL CALLBACK EnumJoysticksCallback(const DIDEVICEINSTANCE*     
                                       pdidInstance, VOID* pContext)
{
    //obtain an interface to the enumerated joystick
    HRESULT hr = inputInstance.m_pDI->CreateDevice(pdidInstance->guidInstance,  
            &inputInstance.m_pdiJoysticks[inputInstance.m_iJoystickCount],
            NULL);

    if(FAILED(hr)) 
        return DIENUM_CONTINUE;

    //increment our joystick counter
    inputInstance.m_iJoystickCount++;
    if (inputInstance.m_iJoystickCount < 10)
        return DIENUM_CONTINUE;

    return DIENUM_STOP;
}

INT32 initDirectInput()
{
    if (FAILED(DirectInput8Create(GetModuleHandle(NULL), DIRECTINPUT_VERSION, 
            IID_IDirectInput8, (void**)&inputInstance.m_pDI, NULL)))
    { 
        return -1;
    }

    //create the keyboard device
    if (FAILED(inputInstance.m_pDI->CreateDevice(GUID_SysKeyboard,
            &inputInstance.m_pdiKeyboard, NULL)))
    { 
        return -1; 
    } 

    //set the keyboard data format
    if (FAILED(inputInstance.m_pdiKeyboard->SetDataFormat(&c_dfDIKeyboard))) {
        return -1; 
    }

    //enumerate and initialize the joysticks
    inputInstance.m_iJoystickCount = 0;
    inputInstance.m_pDI->EnumDevices(DI8DEVCLASS_GAMECTRL,
            EnumJoysticksCallback, NULL, DIEDFL_ATTACHEDONLY);
 
    //get the names of the joysticks
    for (INT32 i = 0; i < inputInstance.m_iJoystickCount; i++) {
        DIPROPSTRING stringProp;
        stringProp.diph.dwSize = sizeof(DIPROPSTRING);
        stringProp.diph.dwHeaderSize = sizeof(DIPROPHEADER);
        stringProp.diph.dwObj = 0;
        stringProp.diph.dwHow = DIPH_DEVICE;
        if (SUCCEEDED(inputInstance.m_pdiJoysticks[i]->
                GetProperty(DIPROP_INSTANCENAME, &stringProp.diph)))
        {
            WideCharToMultiByte(CP_ACP, 0, stringProp.wsz, -1,
                    inputInstance.m_JoystickNames[i], MAX_PATH, NULL, NULL);
        }
        else
            inputInstance.m_JoystickNames[i][0] = 0;
    }

    return 0;
}

INT32 initDirectInputDevices(HWND hWnd)
{
    //set the keyboard cooperative level
    if (FAILED(inputInstance.m_pdiKeyboard->SetCooperativeLevel(
            hWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE)))
        return -1; 

    INT32 j;
    for (INT32 i = 0; i < inputInstance.m_iJoystickCount; i++) {
        if (FAILED(inputInstance.m_pdiJoysticks[i]->SetDataFormat(
                &c_dfDIJoystick2)))
            return -1;

        if (FAILED(inputInstance.m_pdiJoysticks[i]->SetCooperativeLevel(
                hWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE)))
            return -1;

        //set up the dead zone for the axes
        DIPROPDWORD wordProp;
        wordProp.diph.dwSize = sizeof(DIPROPDWORD);
        wordProp.diph.dwHeaderSize = sizeof(DIPROPHEADER);
        wordProp.diph.dwObj = 0;
        wordProp.diph.dwHow = DIPH_DEVICE;
        wordProp.dwData = 1000;
        inputInstance.m_pdiJoysticks[i]->SetProperty(DIPROP_DEADZONE,
                &wordProp.diph);

        //poll the joysticks once so we can determine the center positions
        if (FAILED(inputInstance.m_pdiJoysticks[i]->Acquire()))
            return -1;
    
        if (FAILED(inputInstance.m_pdiJoysticks[i]->Poll()))
            return -1;

        if (FAILED(inputInstance.m_pdiJoysticks[i]->GetDeviceState(
                sizeof(DIJOYSTATE2), &inputInstance.m_JoystickStates[i])))
            return -1;
        
        //get the ranges and centering info for the axes
        DIPROPRANGE rangeProp;
        rangeProp.diph.dwSize = sizeof(DIPROPRANGE);
        rangeProp.diph.dwHeaderSize = sizeof(DIPROPHEADER);
        rangeProp.diph.dwHow = DIPH_BYOFFSET;
        for (j = 0; j < 3; j++) {
            rangeProp.diph.dwObj = (j == 0 ? DIJOFS_X :
                    (j == 1 ? DIJOFS_Y : DIJOFS_Z));
            if (SUCCEEDED(inputInstance.m_pdiJoysticks[i]->GetProperty(
                        DIPROP_RANGE, &rangeProp.diph)))
            {
                ((LONG*)&inputInstance.m_JoystickAxes[i])[j*2] = rangeProp.lMax-
                        ((LONG*)&inputInstance.m_JoystickStates[i])[j];
                ((LONG*)&inputInstance.m_JoystickAxes[i])[(j*2)+1] =
                        ((LONG*)&inputInstance.m_JoystickStates[i])[j];
            }
            else {
                ((LONG*)&inputInstance.m_JoystickAxes[i])[j*2] = 0;
                ((LONG*)&inputInstance.m_JoystickAxes[i])[(j*2)+1] = 0;
            }
        }
    }

    return 0;
}

void releaseDirectInput() {
    for (INT32 i = 0; i < inputInstance.m_iJoystickCount; i++) {
        if (inputInstance.m_pdiJoysticks[i] != NULL) {
            inputInstance.m_pdiJoysticks[i]->Unacquire();
            inputInstance.m_pdiJoysticks[i]->Release();
            inputInstance.m_pdiJoysticks[i] = NULL;
        }
    }
    if (inputInstance.m_pdiKeyboard != NULL) {
        inputInstance.m_pdiKeyboard->Unacquire();
        inputInstance.m_pdiKeyboard->Release();
        inputInstance.m_pdiKeyboard = NULL;
    }
    if (inputInstance.m_pDI != NULL) {
        inputInstance.m_pDI->Release();
        inputInstance.m_pDI = NULL;
    }
}

void DirectInputDevice::release(Intellivision* inty)
{
    releaseDirectInput();
}

void DirectInputDevice::getDefaultConfiguration(INT32 deviceID, INT32 controlID,
        InputSignal** signal)
{
    switch (deviceID) {
        case 0:
            (*signal) = DEFAULT_INPUT_SIGNALS_PLAYER_ONE[controlID];
            return;
        case 1:
            (*signal) = DEFAULT_INPUT_SIGNALS_PLAYER_TWO[controlID];
            return;
        case 2:
            (*signal) = DEFAULT_INPUT_SIGNALS_ECS_KEYBOARD[controlID];
            return;
        case 3:
            (*signal) = DEFAULT_INPUT_SIGNALS_DIRECTOR[controlID];
            return;
    }
}

/*
JNIEXPORT void JNICALL Java_com_bliss_emu_devices_win32_DirectInputDevice_initConfigMode(JNIEnv *env, jobject thisObj)
{
    initDirectInput();
    initDirectInputDevices(GetForegroundWindow());
}

jobject checkKeyboardForInputSignal(JNIEnv *env, jobject thisObj)
{
    //check for key presses on the keyboard
    for (INT32 i = 0; i < 256; i++) {
        if (inputInstance.m_KeyboardBuffer[i] & 0x80) {
            DIPROPSTRING propString;
            propString.diph.dwSize = sizeof(DIPROPSTRING);
            propString.diph.dwHeaderSize = sizeof(DIPROPHEADER);
            propString.diph.dwObj = i;
            propString.diph.dwHow = DIPH_BYOFFSET;

            jstring description;
            if (SUCCEEDED(inputInstance.m_pdiKeyboard->GetProperty(
                    DIPROP_KEYNAME, &propString.diph)))
                description = env->NewString(propString.wsz,
                        lstrlenW(propString.wsz));
            else
                description = env->NewStringUTF("<???>");
            return env->NewObject(g_DirectInputSignalClass,
                    g_DirectInputSignalConstructor, 0, i, description);
        }
    }

    return NULL;
}

jobject checkJoystickForInputSignal(JNIEnv *env, jobject thisObj, INT32 joystickNum)
{
    if (joystickNum >= inputInstance.m_iJoystickCount)
        return NULL;

    DIJOYSTATE2* joyState = &inputInstance.m_JoystickStates[joystickNum];
    //poll the axes
    if (joyState->lX < inputInstance.m_JoystickAxes[joystickNum].lCenterX) {
        jstring description;
        DIDEVICEOBJECTINSTANCE deviceInfo;
        deviceInfo.dwSize = sizeof(DIDEVICEOBJECTINSTANCE);
        if (SUCCEEDED(inputInstance.m_pdiJoysticks[joystickNum]->
                GetObjectInfo(&deviceInfo, DIJOFS_X, DIPH_BYOFFSET)))
        {
            char buf[MAX_PATH];
            sprintf(buf, "-%s (%s)", deviceInfo.tszName,
                    inputInstance.m_JoystickNames[joystickNum]);
            description = env->NewStringUTF(buf);
        }
        else
            description = env->NewStringUTF("<???>");
        return env->NewObject(g_DirectInputSignalClass,
                g_DirectInputSignalConstructor, joystickNum+1, 0, description);
    }

    if (joyState->lY < inputInstance.m_JoystickAxes[joystickNum].lCenterY) {
        jstring description;
        DIDEVICEOBJECTINSTANCE deviceInfo;
        deviceInfo.dwSize = sizeof(DIDEVICEOBJECTINSTANCE);
        if (SUCCEEDED(inputInstance.m_pdiJoysticks[joystickNum]->
                GetObjectInfo(&deviceInfo, DIJOFS_Y, DIPH_BYOFFSET)))
        {
            char buf[MAX_PATH];
            sprintf(buf, "-%s (%s)", deviceInfo.tszName,
                    inputInstance.m_JoystickNames[joystickNum]);
            description = env->NewStringUTF(buf);
        }
        else
            description = env->NewStringUTF("<???>");
        return env->NewObject(g_DirectInputSignalClass,
                g_DirectInputSignalConstructor, joystickNum+1, 1, description);
    }

    if (joyState->lZ < inputInstance.m_JoystickAxes[joystickNum].lCenterZ) {
        jstring description;
        DIDEVICEOBJECTINSTANCE deviceInfo;
        deviceInfo.dwSize = sizeof(DIDEVICEOBJECTINSTANCE);
        if (SUCCEEDED(inputInstance.m_pdiJoysticks[joystickNum]->
                GetObjectInfo(&deviceInfo, DIJOFS_Z, DIPH_BYOFFSET)))
        {
            char buf[MAX_PATH];
            sprintf(buf, "-%s (%s)", deviceInfo.tszName,
                    inputInstance.m_JoystickNames[joystickNum]);
            description = env->NewStringUTF(buf);
        }
        else
            description = env->NewStringUTF("<???>");
        return env->NewObject(g_DirectInputSignalClass,
                g_DirectInputSignalConstructor, joystickNum+1, 2, description);
    }

    if (joyState->lX > inputInstance.m_JoystickAxes[joystickNum].lCenterX) {
        jstring description;
        DIDEVICEOBJECTINSTANCE deviceInfo;
        deviceInfo.dwSize = sizeof(DIDEVICEOBJECTINSTANCE);
        if (SUCCEEDED(inputInstance.m_pdiJoysticks[joystickNum]->
                GetObjectInfo(&deviceInfo, DIJOFS_X, DIPH_BYOFFSET)))
        {
            char buf[MAX_PATH];
            sprintf(buf, "+%s (%s)", deviceInfo.tszName,
                    inputInstance.m_JoystickNames[joystickNum]);
            description = env->NewStringUTF(buf);
        }
        else
            description = env->NewStringUTF("<???>");
        return env->NewObject(g_DirectInputSignalClass,
                g_DirectInputSignalConstructor, joystickNum+1, 3, description);
    }

    if (joyState->lY > inputInstance.m_JoystickAxes[joystickNum].lCenterY) {
        jstring description;
        DIDEVICEOBJECTINSTANCE deviceInfo;
        deviceInfo.dwSize = sizeof(DIDEVICEOBJECTINSTANCE);
        if (SUCCEEDED(inputInstance.m_pdiJoysticks[joystickNum]->
                GetObjectInfo(&deviceInfo, DIJOFS_Y, DIPH_BYOFFSET)))
        {
            char buf[MAX_PATH];
            sprintf(buf, "+%s (%s)", deviceInfo.tszName,
                    inputInstance.m_JoystickNames[joystickNum]);
            description = env->NewStringUTF(buf);
        }
        else
            description = env->NewStringUTF("<???>");
        return env->NewObject(g_DirectInputSignalClass,
                g_DirectInputSignalConstructor, joystickNum+1, 4, description);
    }

    if (joyState->lZ > inputInstance.m_JoystickAxes[joystickNum].lCenterZ) {
        jstring description;
        DIDEVICEOBJECTINSTANCE deviceInfo;
        deviceInfo.dwSize = sizeof(DIDEVICEOBJECTINSTANCE);
        if (SUCCEEDED(inputInstance.m_pdiJoysticks[joystickNum]->
                GetObjectInfo(&deviceInfo, DIJOFS_Z, DIPH_BYOFFSET)))
        {
            char buf[MAX_PATH];
            sprintf(buf, "+%s (%s)", deviceInfo.tszName,
                    inputInstance.m_JoystickNames[joystickNum]);
            description = env->NewStringUTF(buf);
        }
        else
            description = env->NewStringUTF("<???>");
        return env->NewObject(g_DirectInputSignalClass,
                g_DirectInputSignalConstructor, joystickNum+1, 5, description);
    }

    //poll the joystick buttons
    for (INT32 i = 0; i < 128; i++) {
        if (joyState->rgbButtons[i] & 0x80) {
            jstring description;
            DIDEVICEOBJECTINSTANCE deviceInfo;
            deviceInfo.dwSize = sizeof(DIDEVICEOBJECTINSTANCE);
            if (SUCCEEDED(inputInstance.m_pdiJoysticks[joystickNum]->
                    GetObjectInfo(&deviceInfo, DIJOFS_BUTTON(i),
                        DIPH_BYOFFSET)))
            {
                char buf[MAX_PATH];
                sprintf(buf, "%s (%s)", deviceInfo.tszName,
                        inputInstance.m_JoystickNames[joystickNum]);
                description = env->NewStringUTF(buf);
            }
            else
                description = env->NewStringUTF("<???>");
            return env->NewObject(g_DirectInputSignalClass,
                    g_DirectInputSignalConstructor, joystickNum+1, i+6,
                    description);
        }
    }

    //poll the POVs
    for (i = 0; i < 4; i++) {
        if (LOWORD(inputInstance.m_JoystickStates[joystickNum].rgdwPOV[i]) ==
                0xFFFF)
            continue;

        //determine the direction the user moved the POV
        INT32 povValue = inputInstance.m_JoystickStates[joystickNum].rgdwPOV[i];
        INT32 direction = ((povValue + 4500) / 9000) & 0x03;

        jstring description;
        DIDEVICEOBJECTINSTANCE deviceInfo;
        deviceInfo.dwSize = sizeof(DIDEVICEOBJECTINSTANCE);
        if (SUCCEEDED(inputInstance.m_pdiJoysticks[joystickNum]->
                GetObjectInfo(&deviceInfo, DIJOFS_POV(i),
                    DIPH_BYOFFSET)))
        {
            jclass clazz = env->FindClass(
                    "com/bliss/emu/devices/win32/Win32PlugIn");
            jfieldID resourceField = env->GetStaticFieldID(clazz, "RESOURCES",
                    "Ljava/util/ResourceBundle;");
            jobject resources = env->GetStaticObjectField(clazz, resourceField);
            jmethodID method = env->GetMethodID(env->GetObjectClass(resources),
                    "getString", "(Ljava/lang/String;)Ljava/lang/String;");
            jstring key = env->NewStringUTF(
                    g_lpszPOVDescriptionKeys[direction]);
            jstring theString = (jstring)env->CallObjectMethod(resources,
                    method, key);
            const char* direction = env->GetStringUTFChars(theString, NULL);
            char buf[MAX_PATH];
            sprintf(buf, "%s %s (%s)", deviceInfo.tszName, direction,
                    inputInstance.m_JoystickNames[joystickNum]);
            env->ReleaseStringUTFChars(theString, direction);
            description = env->NewStringUTF(buf);
        }
        else
            description = env->NewStringUTF("<???>");
        return env->NewObject(g_DirectInputSignalClass,
                    g_DirectInputSignalConstructor, joystickNum+1,
                    262+(i*4)+direction,
                    description);
    }

    return NULL;
}

JNIEXPORT jobject JNICALL Java_com_bliss_emu_devices_win32_DirectInputDevice_pollForInputSignal(JNIEnv *env, jobject thisObj)
{
    if (!poll())
        return NULL;

    jobject returnValue = checkKeyboardForInputSignal(env, thisObj);
    if (returnValue != NULL)
        return returnValue;

    for (INT32 i = 0; i < inputInstance.m_iJoystickCount; i++) {
        returnValue = checkJoystickForInputSignal(env, thisObj, i);
        if (returnValue != NULL)
            return returnValue;
    }

    return NULL;
}

JNIEXPORT void JNICALL Java_com_bliss_emu_devices_win32_DirectInputDevice_releaseConfigMode(JNIEnv *env, jobject thisObj)
{
    releaseDirectInput();
}
*/

INT32 DirectInputDevice::init(Intellivision* inty)
{
    INT32 result;
    if (result = initDirectInput()) {
        return result;
    }

    if (result = initDirectInputDevices(currentDevice->m_hWnd)) {
        releaseDirectInput();
        return result;
    }

    return 0;
}

float getJoystickSignalValue(INT32 joystickNum, INT32 inputCode)
{
    if (joystickNum >= inputInstance.m_iJoystickCount)
        return 0.0f;

    if (inputCode < 6) {
        LONG inputValue = 0;
        LONG maxValue = 0;
        switch (inputCode) {
            case 0:
                inputValue = -(inputInstance.m_JoystickStates[joystickNum].lX -
                        inputInstance.m_JoystickAxes[joystickNum].lCenterX);
                maxValue = inputInstance.m_JoystickAxes[joystickNum].lMaxDeltaX;
                break;
            case 1:
                inputValue = -(inputInstance.m_JoystickStates[joystickNum].lY -
                        inputInstance.m_JoystickAxes[joystickNum].lCenterY);
                maxValue = inputInstance.m_JoystickAxes[joystickNum].lMaxDeltaY;
                break;
            case 2:
                inputValue = -(inputInstance.m_JoystickStates[joystickNum].lZ -
                        inputInstance.m_JoystickAxes[joystickNum].lCenterZ);
                maxValue = inputInstance.m_JoystickAxes[joystickNum].lMaxDeltaZ;
                break;
            case 3:
                inputValue = inputInstance.m_JoystickStates[joystickNum].lX -
                        inputInstance.m_JoystickAxes[joystickNum].lCenterX;
                maxValue = inputInstance.m_JoystickAxes[joystickNum].lMaxDeltaX;
                break;
            case 4:
                inputValue = inputInstance.m_JoystickStates[joystickNum].lY -
                        inputInstance.m_JoystickAxes[joystickNum].lCenterY;
                maxValue = inputInstance.m_JoystickAxes[joystickNum].lMaxDeltaY;
                break;
            case 5:
                inputValue = inputInstance.m_JoystickStates[joystickNum].lZ -
                        inputInstance.m_JoystickAxes[joystickNum].lCenterZ;
                maxValue = inputInstance.m_JoystickAxes[joystickNum].lMaxDeltaZ;
                break;
        }

        if (inputValue <= 0)
            return 0.0f;

        return (float)((float)inputValue)/((float)maxValue);
    }
    else if (inputCode < 262) {
        return (inputInstance.m_JoystickStates[joystickNum].
                rgbButtons[inputCode-6] & 0x80 ? 1.0f : 0.0f);
    }
    else if (inputCode < 278) {
        INT32 directionToCheck = (inputCode-262) & 0x03;
        DWORD povValue = inputInstance.m_JoystickStates[joystickNum].
                rgdwPOV[((inputCode-262) & 0x0C) >> 2];
        if (LOWORD(povValue) == 0xFFFF)
            return 0.0f;

        INT32 direction = ((povValue + 4500) / 9000) & 0x03;
        return (direction == directionToCheck ? 1.0f : 0.0f);
    }
    return 0.0f;
}

float DirectInputDevice::getSignalValue(InputSignal* inputSignal)
{
    if (!inputInstance.m_bLastPollSuccessful || inputSignal == NULL)
        return 0.0f;

    DirectInputSignal* dinputSignal = (DirectInputSignal*)inputSignal;
    if (dinputSignal->deviceID == 0)
        return (inputInstance.m_KeyboardBuffer[dinputSignal->controlID] & 0x80
                ? 1.0f : 0.0f);
    else
        return getJoystickSignalValue(dinputSignal->deviceID-1,
                dinputSignal->controlID);
    return 0.0f;
}


